package game.command;

import java.lang.reflect.Array;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

import game.command.DoubleParser.DefType;
import game.world.Vec3;
import game.world.World;

public abstract class ScriptExecutable implements Executable {
	private final String name;
	private final Map<String, Parameter> parameters = Maps.newHashMap();
	private final List<Parameter> argList = Lists.newArrayList();
	
	private int parPos = 0;
	private boolean parReq = true;
	
	protected ScriptExecutable(String name) {
		this.name = name;
	}
	
	protected ScriptExecutable setParamsOptional() {
		this.parReq = false;
		return this;
	}
	
	protected ScriptExecutable addParameter(String name, ArgCombiner<?> combiner, ArgumentParser ... parsers) {
		Parameter param = new Parameter(name, (char)0, this.parPos++, this.parReq, Lists.newArrayList(parsers), combiner);
		this.parameters.put(name, param);
		this.argList.add(param);
		return this;
	}
	
	protected ScriptExecutable addParameter(String name, char shortName, ArgCombiner<?> combiner, ArgumentParser ... parsers) {
		Parameter param = new Parameter(name, shortName, -1, false, Lists.newArrayList(parsers), combiner);
		this.parameters.put(name, param);
		this.argList.add(param);
		return this;
	}
	
	protected ScriptExecutable addParameter(String name, ArgumentParser ... parsers) {
		return this.addParameter(name, null, parsers);
	}
	
	protected ScriptExecutable addParameter(String name, char shortName, ArgumentParser ... parsers) {
		return this.addParameter(name, shortName, null, parsers);
	}
	
	protected ScriptExecutable addParameter(ArgumentParser parser) {
		return this.addParameter(parser.getName(), parser);
	}
	
	protected ScriptExecutable addParameter(char shortName, ArgumentParser parser) {
		return this.addParameter(parser.getName(), shortName, parser);
	}
	
	protected ScriptExecutable addVector(String name, boolean defaulted) {
		return this.addParameter(name, new ArgCombiner<Double>() {
			public Vec3 combine(Double[] values) {
				return new Vec3(values[0], values[1], values[2]);
			}
			
			public Class<?> getTypeClass() {
				return Vec3.class;
			}
			
			public Class<Double> getInputClass() {
				return Double.class;
			}
		},
		new DoubleParser("x", defaulted ? DefType.X : null, (double)(-World.MAX_SIZE), (double)World.MAX_SIZE), 
		new DoubleParser("y", defaulted ? DefType.Y : null, 0.0, (double)World.HEIGHT),
		new DoubleParser("z", defaulted ? DefType.Z : null, (double)(-World.MAX_SIZE), (double)World.MAX_SIZE));
	}
	
	protected ScriptExecutable addWorld(String name, boolean defaulted) {
		return this.addParameter(new WorldParser(name, false, defaulted));
	}
	
	protected ScriptExecutable addInt(String name, char shortName, int min, int max, int def) {
		return this.addParameter(shortName, new IntParser(name, false, def, min, max, def));
	}
	
	protected ScriptExecutable addInt(String name, int min, int max, int def) {
		return this.addParameter(new IntParser(name, false, def, min, max, def));
	}
	
	protected ScriptExecutable addFlag(String name, char shortName) {
		return this.addParameter(name, shortName);
	}
	
//	protected ScriptExecutable addFlag(String name) {
//		return this.addParameter(name, name.charAt(0));
//	}
	
	protected <T> ScriptExecutable addEnum(String name, T def, Class<T> clazz, T ... values) {
		return this.addParameter(new EnumParser<T>(name, clazz, def, values));
	}
	
	protected <T> ScriptExecutable addEnum(String name, Class<T> clazz, T ... values) {
		return this.addEnum(name, null, clazz, values);
	}
	
	protected <T> ScriptExecutable addEnum(String name, T def, Class<T> clazz, Collection<T> values) {
		return this.addEnum(name, def, clazz, values.toArray((T[])Array.newInstance(clazz, values.size())));
	}
	
	protected <T> ScriptExecutable addEnum(String name, Class<T> clazz, Collection<T> values) {
		return this.addEnum(name, null, clazz, values);
	}
	
	protected ScriptExecutable addTag(String name) {
		return this.addParameter(new TagParser(name, null));
	}
	
	protected ScriptExecutable addTag(String name, char shortName) {
		return this.addParameter(shortName, new TagParser(name, null));
	}
	
//	public Object exec(ScriptEnvironment env, ScriptArgs args) {
//		this.getClass().getDeclaredMethod(name, null)
//		return null;
//	}

	public Map<String, Parameter> getParameters() {
		return this.parameters;
	}

	public List<Parameter> getParamList() {
		return this.argList;
	}

	public String getName() {
		return this.name;
	}
}
